 /**********************************************
 * This program for the atmega168 works like   *
 * an interpreter, where the microcontroller   *
 * sends and receives data and instructions    *
 * from a PC over a serial interface.          *
 * The primary function of this implementation *
 * is the control of the built-in analog to    *
 * digital converter (ADC)                     *
 *                                             *
 * Relevent documentation                      *
 * Atmel ATmega48/88/168 data sheet            *
 *                                             *
 * Communication protocol:                     *
 *                                             *
 * Communication is over 57600 baud serial     *
 * with 8 data bits, 1 stop bit and, no parity *
 *                                             *
 * The device accepts one byte commands. When  *
 * a command and is received, it is processed  *
 * immediately. If another command is issued   *
 * to the device while the previous is still   * 
 * being processed, it will be ignored.        *
 *                                             *
 * The device will ignore unrecognized         *
 * commands.                                   *
 *                                             *
 * List of Commands:                           *
 * Req. = # of request bytes                   *
 * Res. = # of response bytes                  *
 * ------------------------------------------- *
 * Hex   Ascii  Req.  Res.  Description        *
 * ------------------------------------------- *
 * 0x4C  'L'    0     0     toggle status LED  *
 * 0x41  'C'    0     -     adc - start conv.  *
 * 0x48  'H'    0     0     adc - halt conv.   *
 **********************************************/

/**********************************************************/
/* This program is free software; you can redistribute it */
/* and/or modify it under the terms of the GNU General    */
/* Public License as published by the Free Software       */
/* Foundation; either version 2 of the License, or        */
/* (at your option) any later version.                    */
/*                                                        */
/* This program is distributed in the hope that it will   */
/* be useful, but WITHOUT ANY WARRANTY; without even the  */
/* implied warranty of MERCHANTABILITY or FITNESS FOR A   */
/* PARTICULAR PURPOSE.  See the GNU General Public        */
/* License for more details.                              */
/*                                                        */
/* You should have received a copy of the GNU General     */
/* Public License along with this program; if not, write  */
/* to the Free Software Foundation, Inc.,                 */
/* 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA */
/*                                                        */
/* Licence can be viewed at                               */
/* http://www.fsf.org/licenses/gpl.txt                    */
/**********************************************************/

#include <avr/io.h>
#include <avr/interrupt.h>

#define FOSC F_CPU
#define BAUD 57600
#define MYUBRR (FOSC/16/BAUD-1)

#define sbi(var, mask)   ((var) |= (uint8_t)(1 << mask))
#define cbi(var, mask)   ((var) &= (uint8_t)~(1 << mask))

//Define delay loop values
//=====================
#define DELAY_500us_8MHz_16bit 0x03E5 // 997
#define DELAY_70us_8MHz_16bit  0x008A // 138
#define DELAY_60us_8MHz_16bit  0x0075 // 117

#define DELAY_70us_8MHz_8bit   0xB9   // 185
#define DELAY_60us_8MHz_8bit   0x9E   // 158
#define DELAY_20us_8MHz_8bit   0x34   // 52
#define DELAY_15us_8MHz_8bit   0x25   // 38
#define DELAY_10us_8MHz_8bit   0x19   // 25
#define DELAY_5us_8MHz_8bit    0x0C   // 12
#define DELAY_2us_8MHz_8bit    0x04   // 4
#define DELAY_1us_8MHz_8bit    0x01   // 1
//=====================

//Define globals
//=====================
volatile uint8_t  dct8;           //the 8-bit delay counter, used in asm
volatile uint16_t dct16;          //the 16-bit delay counter, used in asm
volatile uint8_t  ch;             //the character received over USART
//=====================

//Define functions and macros
//======================
static void    io_init(void);

static void    usart_tx(uint8_t);
static uint8_t usart_rx(void);
static void    toggle_LED(void);
static void    delay_ms(uint16_t);

static void    adc_conv(void);
static void    adc_halt(void);

/* The delay8() macro uses (dct8 * 3 + 5) cycles.
 * The global variable dct8 must be set before
 * the macro is used. Since dct is an uint8_t,
 * the macro can only loop a maximum of
 * 2^8 = 256 times. This is equivalent to
 * 96.625us maximum, and 0.3125us minimum. */
#define delay8()\
  asm volatile(\
    "push  r24          \n\t"\
    "lds   r24, dct8    \n\t"\
    "1:                 \n\t"\
    "dec   r24          \n\t"\
    "brne  1b           \n\t"\
    "pop   r24          \n\t"\
  )

/* The delay16() macro uses (dct16 * 4 + 11) cycles.
 * The global variable dct16 must be set before
 * the macro is used. Since dct is an uint16_t,
 * the macro can only loop a maximum of
 * 2^16 = 65536 times. This is equivalent to
 * 32769.38us maximum, and 1.375us minimum. */
#define delay16()\
  asm volatile(\
    "push  r24          \n\t"\
    "push  r25          \n\t"\
    "lds   r24, dct16   \n\t"\
    "lds   r25, dct16+1 \n\t"\
    "1:                 \n\t"\
    "sbiw  r24, 1       \n\t"\
    "brne  1b           \n\t"\
    "pop   r25          \n\t"\
    "pop   r24          \n\t"\
  )
//======================

//Define ISR
//======================
ISR(USART_RX_vect) {
  ch = (uint8_t) usart_rx();
  switch( ch ) {
    case 'L' :
      toggle_LED();
      break;
    case 'C' :
      adc_conv();
      break;
    case 'H' :
      adc_halt(); 
      break;
    default  :
      break;
  }
}

ISR(TIMER1_COMPA_vect) {
  usart_tx(ADCL);
  usart_tx(ADCH);
}
//======================

//Function definitions
//======================
int main (void) {
  io_init();
  while(1) {}
  return 0;
}

void io_init (void){
  DDRC = 0x00; // configure PORTC as input (0) for status LED
  cbi(PORTC, PORTC5); //tri-state the status LED

  // Set up USART0
  UBRR0H = (MYUBRR>>8); //set baudrate high byte
  UBRR0L = MYUBRR; //set baudrate low byte
  UCSR0C = (1<<UCSZ00)|(1<<UCSZ01); //async 8 bit, no parity, 1 stop bit
  UCSR0B = (1<<RXCIE0)|(1<<RXEN0)|(1<<TXEN0); //enable rx, tx, rx int

  sei(); //enable Global interrupts
}

static void usart_tx(uint8_t c) {
  while( !(UCSR0A & (1<<UDRE0)) );
  UDR0 = c;
}

static uint8_t usart_rx() {
    while( !(UCSR0A & (1<<RXC0)) );
    return(UDR0);
}

static void delay_ms( uint16_t ms ) {
  ms /= 32;
  dct16 = 0;
  while( ms-- > 0 ) { delay16(); }
}

static void toggle_LED(void) {
  DDRC ^= (1<<DDC5);
}

static void adc_conv(void) {
  //enable ADC, auto-trigger, 128 clock division
  ADCSRA = (1<<ADEN)|(1<<ADATE)|(1<<ADPS0)|(1<<ADPS1)|(1<<ADPS2);
  //select voltage reference
  ADMUX = (1<<REFS0);
  //set TOP (OCR1A) = F_CPU / 1000 
  OCR1A = (uint16_t)(F_CPU / 1000);
  //enable output compare A interrupts
  TIMSK1 = (1<<OCIE1A);                                                                             
  //select clear timer on compare (CTC) mode
  TCCR1B = (1<<WGM12);
  //start ADC conversion
  ADCSRA |= (1<<ADSC);
  //select counter clock source 1 (no prescaler), start counter
  TCCR1B |= (1<<CS10);
}

static void adc_halt(void) {
  //disable counter, reset configuration
  TCCR1B = 0;
  TIMSK1 = 0;
  OCR1A  = 0x0000;
  //disable ADC, reset configuration
  ADCSRA = 0;
  ADMUX  = 0;
}

//======================
